{
    Driver->Storage->DriveManager Physical storage device manager

    @author(Aaron Hance ah@aaronhance.me)
}
unit drivemanager;

interface

uses
    util,
    drivertypes,
    console,
    terminal,
    drivermanagement,
    lmemorymanager,
    strings,
    lists,
    tracer,
    rtc,
    MBR,
    serial;

type 
    TControllerType = (ControllerIDE, ControllerUSB, ControllerAHCI, 
        ControllerNET, ControllerRAM, rsvctr1, rsvctr2, rsvctr3);

    PStorage_device = ^TStorage_Device;
    PDrive_Error    = ^TDrive_Error;

    // PPWriteHook = procedure(volume : PStorage_volume; directory : pchar; entry : PDirectory_Entry; byteCount : uint32; buffer : puint32; statusOut : puint32);
    // PPReadHook = function(volume : PStorage_Volume; directory : pchar; fileName : pchar; fileExtension : pchar; buffer : puint32; bytecount : puint32) : uint32;
    // PPCreateHook = procedure(disk : PStorage_Device; sectors : uint32; start : uint32; config : puint32);
    // PPDetectHook = procedure(disk : PStorage_Device);
    // PPCreateDirHook = procedure(volume : PStorage_volume; directory : pchar; dirname : pchar; attributes : uint32; status : puint32);
    // PPReadDirHook = function(volume : PStorage_volume; directory : pchar; status : puint32) : PLinkedListBase; //returns: 0 = success, 1 = dir not exsist, 2 = not directory, 3 = error //returns: 0 = success, 1 = dir not exsist, 2 = not directory, 3 = error
    PPHIOHook = procedure(drive : PStorage_device; addr : uint32; sectors : uint32; buffer : puint32);

    PPHIOHook_ = procedure;


    { Generic storage device }
    TStorage_Device = record
        id               : uint8;
        controller       : TControllerType;
        controllerId0    : uint32;
        writable         : boolean;
        volumes          : PLinkedListBase;
        writeCallback    : PPHIOHook;
        readCallback     : PPHIOHook;
        maxSectorCount   : uint32;
        sectorSize       : uint32;
    end;

    TDrive_Error = record 
        code : uint16;
        description : pchar;
        recoverable : boolean;
    end;

var 
    storageDevices : PLinkedListBase; //index in this array is global drive id

    procedure init();

    procedure register_device(device : PStorage_Device);
    function get_drive_list() : PLinkedListBase;
    function controller_type_2_string(controllerType : TControllerType) : pchar; 

implementation 

procedure list_drive_command();
var 
    i : uint32 = 0;
begin
    push_trace('DriverManager.list_drive_command');
    //if no storage device print none found
    if LL_Size(storageDevices) < 1 then begin
        console.writestringlnWnd('No storage devices found.', getTerminalHWND());
        exit;
    end;

    //print number of storage devices
    console.writeintwnd(LL_Size(storageDevices), getTerminalHWND());
    console.writestringlnWnd(' devices found', getTerminalHWND());

    for i:=0 to LL_Size(storageDevices)-1 do begin

        //print device id and type
        console.writeintwnd(i, getTerminalHWND());
        console.writestringwnd(' - Device type: ', getTerminalHWND());
        console.writestringwnd(controller_type_2_string(PStorage_Device(LL_Get(storageDevices, i))^.controller), getTerminalHWND());
        console.writestringwnd(', ', getTerminalHWND());

        //print device capcity
        console.writestringwnd(' Capacity: ', getTerminalHWND());
        console.writeintwnd((( PStorage_Device(LL_Get(storageDevices, i))^.maxSectorCount * PStorage_Device(LL_Get(storageDevices, i))^.sectorSize) DIV 1024) DIV 1024, getTerminalHWND());
        console.writestringlnWnd('MB', getTerminalHWND());
    end;

end;

{ initalise a drive with a MBR - TODO needs to be expanded for partitioning }
procedure init_drive_command(driveIndex : uint32);
var
    bootrecord : PMaster_Boot_Record;
    partition : PPartition_table;
    drive : PStorage_device;
begin
    push_trace('DriverManager.init_drive_command');
    //setup MBR and partition table
    bootrecord := PMaster_Boot_record(kalloc(SizeOf(TMaster_boot_record)));
    partition := PPartition_table(kalloc(SizeOf(TPartition_table)));

    drive := PStorage_device(LL_Get(storageDevices, driveIndex));

    MBR.setup_partition(partition, 100, drive^.maxSectorCount-300);
    bootrecord^.partition_0 := partition^;

    //write MBR and partition table to disk
    drive^.writeCallback(drive, 0, 1, puint32(bootrecord));

    //volume_manager reload partition
    //TODO

end;

procedure drive_command(params : PParamList);
var
    i : uint16;
    drive : uint16;
    subcmd : pchar;
begin
    push_trace('DriverManager.drive_command');
    subcmd:= getParam(0, params);

    if ((paramCount(params) = 0)) then begin
        console.writestringlnWnd('Please provide valid arguments.', getTerminalHWND());
        console.writestringlnWnd('  ls - for listing all drives', getTerminalHWND());
        console.writestringlnWnd('  info [drive] - display formation for specified drive', getTerminalHWND());
        console.writestringlnWnd('  init [drive] - destructive, formats a drive with a blank partition', getTerminalHWND());
        exit;
    end;

    if( stringEquals(subcmd, 'ls') ) then begin
      list_drive_command();
      exit;
    end;

    if( stringEquals(subcmd, 'info') ) then begin
      //TODO unimplimented
      exit
    end;

    if( stringEquals(subcmd, 'init') ) then begin
      init_drive_command(stringToInt(getParam(1, params) ) );
      exit
    end;
end;

{ Initialise the drive manager }
procedure init();
begin
    push_trace('DriveManager.init');
    
    //setup drive manager
    setworkingdirectory('.');
    storageDevices:= ll_new(sizeof(TStorage_Device));

    serial.sendString('HERE I DID THE INIT WAH DU HEK BRUH!!!!!!');

    //register DISK command 
    terminal.registerCommand('DISK', @drive_command, 'Disk utility');
end;

{ Register a new drive with the drive manager }
procedure register_device(device : PStorage_device);
var 
    i   : uint8;
    elm : void;
begin
    push_trace('DriveManager.register_device');

    //add the drive to the list of storage devices.
    elm:= LL_Add(storageDevices); 
    PStorage_device(elm)^ := device^; //TODO memcopy

    //set drive id in struct
    PStorage_device(elm)^.id := LL_Size(storageDevices) - 1;

    //create empty volume list for drive
    // PStorage_Device(LL_Get(storageDevices, LL_Size(storageDevices) - 1))^.volumes := LL_New(sizeof(TStorage_Volume));
    
    //TODO register with volume manager

end;

{ Get the drive list }
function get_drive_list() : PLinkedListBase;
begin
    push_trace('DriverManager.get_drive_list');

    //return the device list
    get_drive_list:= storageDevices;
end;

{ Get string name for controller type }
function controller_type_2_string(controllerType : TControllerType) : pchar; 
begin
    push_trace('DriverManager.controller_type_2_string');

    //return string of controller type
    case controllerType of
        ControllerIDE : controller_type_2_string:= 'IDE';
        ControllerUSB :  controller_type_2_string:= 'USB';
        ControllerAHCI : controller_type_2_string:= 'ACHI (SATA)';
        ControllerNET : controller_type_2_string:= 'Net';
    end;
end;

function write_2_drive(drive : uint32; address : uint32; byteSize : uint32; buffer : puint32) : PDrive_Error;
begin
  
end;

end.