{
    Driver->Storage->VolumeManager Drive volume manager

    @author(Aaron Hance ah@aaronhance.me)
}
unit volumemanager;

interface

uses
    util, 
    console,
    drivemanager,
    strings,
    lists,
    tracer,
    vfs,
    MBR;

type
    
    TDirectory_Entry_Type = (directoryEntry, fileEntry, mountEntry);

    PDirectory_Entry = ^TDirectory_Entry;
    PFilesystem = ^TFilesystem;
    PStorage_Volume = ^TStorage_Volume;

    PPWriteHook = procedure(volume : PStorage_volume; directory : pchar; entry : PDirectory_Entry; byteCount : uint32; buffer : puint32; statusOut : puint32);
    PPReadHook = function(volume : PStorage_Volume; directory : pchar; fileName : pchar; fileExtension : pchar; buffer : puint32; bytecount : puint32) : uint32;
    PPHIOHook = procedure(volume : PStorage_device; addr : uint32; sectors : uint32; buffer : puint32);
    PPCreateHook = procedure(disk : PStorage_Device; sectors : uint32; start : uint32; config : puint32);
    PPDetectHook = procedure(disk : PStorage_Device);
    PPCreateDirHook = procedure(volume : PStorage_volume; directory : pchar; dirname : pchar; attributes : uint32; status : puint32);
    PPReadDirHook = function(volume : PStorage_volume; directory : pchar; status : puint32) : PLinkedListBase; //returns: 0 = success, 1 = dir not exsist, 2 = not directory, 3 = error //returns: 0 = success, 1 = dir not exsist, 2 = not directory, 3 = error

    PPHIOHook_ = procedure;

    TFilesystem = record
        sName : pchar;
        system_id : uint8;
        writeCallback     : PPWriteHook;
        readCallback      : PPReadHook;
        createCallback    : PPCreateHook;
        detectCallback    : PPDetectHook;
        createDirCallback : PPCreateDirHook;
        readDirCallback   : PPReadDirHook;
    end;

    TStorage_Volume = record
        device          : PStorage_device;
        sectorStart     : uint32;
        sectorSize      : uint32;
        sectorCount     : uint32;
        freeSectors     : uint32;
        filesystem      : PFilesystem; 
        { True if this drive contains the loaded OS }
        isBootDrive     : boolean;
    end;

    { Generic directory entry }
    TDirectory_Entry = record
        { Contains filename and optionally file extension, seperated by the last period.}
        fileName  : pchar; 
        entryType : TDirectory_Entry_Type;
        // creationT : TDateTime;
        // modifiedT : TDateTime;
    end;
  
var

    storageVolumes : PLinkedListBase;
    filesystems : PlinkedListBase;

    procedure init();
    
    procedure register_filesystem(filesystem : PFilesystem);
    procedure check_for_volumes(drive : PStorage_device);

    //IO functions

implementation

{ Initialise volume manager }
procedure init();
begin
    push_trace('VolumeManager.init');

    { setup lists }
    storageVolumes:= ll_New(sizeof(TFilesystem));
    filesystems:= ll_New(sizeof(TFilesystem));      
end;

{ register a new type of filesystem }
procedure register_filesystem(filesystem : PFilesystem); //TODO on init create null filesystem for empty partition
begin
    //add filesystem

    //check drives for volumes of new type
end;

{ Check for volumes on a physical device }
procedure check_for_volumes(drive : PStorage_device);
var 
    bootrecord : PMaster_Boot_Record;
    storageVolume : TStorage_Volume;
    i : uint32 = 0;
    elm : void;
begin

    drive^.readCallback(drive, 0, 1, puint32(bootrecord));

    //TODO multipe partition entries
    if bootrecord^.partition_0.LBA_start <> 0 then
    begin
      //set volume information
      storageVolume.device := drive;
      storageVolume.sectorStart := bootrecord^.partition_0.LBA_start;
      storageVolume.sectorCount := bootrecord^.partition_0.sector_count;
      storageVolume.freeSectors := 0; //TODO impliment

      //check for filesystem type
      for i:=0 to LL_Size(filesystems) - 1 do begin
        if bootrecord^.partition_0.system_id = PFilesystem(LL_Get(filesystems, i))^.system_id then
        begin
          storageVolume.filesystem := PFilesystem(LL_Get(filesystems, i));
        end;
      end;

      //add volume to list
      elm := LL_Add(storageVolumes);
      memcpy(uint32(@storageVolume), uint32(elm), SizeOf(TStorage_Volume));
        
    end;
    
end;

end.